#
#  Copyright ©2024. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#
import base64

import pytest
import requests

from levo_commons.models import Response
from levo_commons.request_utils import (
    get_error_code_from_response_body,
    is_4xx_error_response,
    is_valid_2xx_response,
)


def test_is_4xx_error_response():
    response = Response(
        status_code=400,
        body="{}",
        headers={},
        method="GET",
        uri="/test",
        message="Bad Request",
        http_version="HTTP/1.1",
        encoding=None,
    )
    response.status_code = 400
    assert is_4xx_error_response(response) is True

    # Add json body to the response and verify
    response.body = '{"error": "Bad Request"}'
    assert is_4xx_error_response(response) is True

    # Set status code to 200 but indicate error in the body
    response.status_code = 200
    response.body = '{"error": {"code": 400, "message": "Bad Request"}}'
    assert is_4xx_error_response(response) is True

    response.body = '{"status_code": 400, "message": "Bad Request"}'
    assert is_4xx_error_response(response) is True

    response.status_code = 500
    assert is_4xx_error_response(response) is False


def test_is_valid_2xx_response():
    response = Response(
        status_code=200,
        body="{}",
        headers={},
        method="GET",
        uri="/test",
        message="OK",
        http_version="HTTP/1.1",
        encoding=None,
    )
    assert is_valid_2xx_response(response) is True

    response.status_code = 201
    assert is_valid_2xx_response(response) is True

    response.status_code = 204
    assert is_valid_2xx_response(response) is True

    response.status_code = 400
    assert is_valid_2xx_response(response) is False

    response.status_code = 500
    assert is_valid_2xx_response(response) is False

    response.status_code = 200
    response.body = '{"errorCode": 400, "errorMessage": "Bad Request"}'
    assert is_valid_2xx_response(response) is False

    response.body = '{"statusCode": 400, "errorMessage": "Bad Request"}'
    assert is_valid_2xx_response(response) is False

    response.body = '{"error": {"statusCode": 400, "errorMessage": "Bad Request"}}'
    assert is_valid_2xx_response(response) is False

    response.body = base64.b64encode(
        '{"error": {"statusCode": 400, "errorMessage": "Bad Request"}}'.encode()
    )
    assert is_valid_2xx_response(response) is False


def test_string_error_code():
    response = requests.Response()
    response._content = '{"success":false,"code":"unauthorized","data":null,"msg":"Authorization failed"}'
    assert get_error_code_from_response_body(response) == 401

    response._content = '{"success":false,"code":"invalid_input","data":null,"msg":"Invalid auth token provided"}'
    assert get_error_code_from_response_body(response) == 400

    response._content = (
        '{"success":false,"code":"RETRY","data":{"msg":"Please try again."},"msg":""}'
    )
    assert get_error_code_from_response_body(response) == 503

    response._content = (
        '{"success":true,"code":"OK","data":{"IS_FUTURE_ENABLED":false},"msg":""}'
    )
    assert get_error_code_from_response_body(response) is None

    response._content = '{"success":true,"code":"200","data":{"express_kyc_status":"NOT_STARTED"},"msg":""}'
    assert get_error_code_from_response_body(response) == 200

    response._content = (
        '{"success":false,"status":"retry","error":{"code":401},"data":{'
        '"express_kyc_status":"NOT_STARTED"},"msg":""}'
    )
    assert get_error_code_from_response_body(response) == 401

    response._content = '{"success":true,"code":"401","data":{"express_kyc_status":"NOT_STARTED"},"msg":""}'
    assert get_error_code_from_response_body(response) == 401

    response._content = '{"result":41406, "data":{"error":"wrong_params"},"msg":""}'
    assert get_error_code_from_response_body(response) == 400

    response._content = (
        '{"result":41406, "code":"403", "data":{"error":"invalid_token"},"msg":""}'
    )
    assert get_error_code_from_response_body(response) == 403

    response._content = '{"result":41406, "data":{"error":"invalid_token"},"msg":""}'
    assert get_error_code_from_response_body(response) == 401


@pytest.mark.parametrize("res_body", ["[]", "12345"])
def test_non_dict_json_body(res_body):
    response = requests.Response()
    response._content = res_body
    try:
        assert get_error_code_from_response_body(response) is None
    except AttributeError:
        pytest.fail("Should not raise AttributeError")



#
#  Copyright ©2022. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#

import sys

from hypothesis import given
from hypothesis import strategies as st

from levo_commons.utils import base64_decode, base64_encode, syspath_prepend


def test_syspath_prepend():
    path = "./foo"
    assert path not in sys.path
    with syspath_prepend(path):
        assert sys.path[0] == path
    assert path not in sys.path


def test_syspath_prepend_nested():
    path_1 = "./foo"
    path_2 = "./bar"
    assert path_1 not in sys.path
    assert path_2 not in sys.path
    with syspath_prepend(path_1):
        assert sys.path[0] == path_1
        with syspath_prepend(path_2):
            assert sys.path[0] == path_2
            assert sys.path[1] == path_1
        assert sys.path[0] == path_1
    assert path_1 not in sys.path
    assert path_2 not in sys.path


@given(st.text())
def test_base64_encode_decode(s):
    assert s == base64_decode(base64_encode(s))
    # Test with bytes type
    b = s.encode("utf-8")
    assert b.decode("utf-8") == base64_decode(base64_encode(b))



#
#  Copyright ©2022. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#

from enum import Enum


class ParamType(str, Enum):
    """Resulting status of some action."""

    body = "body"
    query = "query"
    params = "params"
    cookies = "cookies"
    headers = "headers"

    def __str__(self) -> str:
        return self.value



#
#  Copyright ©2022. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#
import enum
from pathlib import Path
from typing import Any, Callable, Dict, List, Set, Tuple, Union

PathLike = Union[Path, str]

Query = Dict[str, Any]
# Body can be of any Python type that corresponds to JSON Schema types + `bytes`
Body = Union[List, Dict[str, Any], str, int, float, bool, bytes]
PathParameters = Dict[str, Any]
Headers = Dict[str, Any]
Cookies = Dict[str, Any]
FormData = Dict[str, Any]


class NotSet:
    pass


# A filter for path / method
Filter = Union[str, List[str], Tuple[str], Set[str], NotSet]

RawAuth = Tuple[str, str]
# Generic test with any arguments and no return
GenericTest = Callable[..., None]


class ResponseValueLocation(enum.Enum):
    BODY = "body"
    HEADER = "header"



#
#  Copyright ©2022. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#

from logging import Handler
from typing import Dict, List, Optional, Tuple

import attr

from .models import Module
from .providers import Provider

CONFIG_VERSION = (1, 0)
DEFAULT_REQUEST_TIMEOUT = 30


@attr.s(slots=True)
class AuthConfig:
    auth_type: str = attr.ib(kw_only=True, default="None")
    username: Optional[str] = attr.ib(kw_only=True, default=None)
    password: Optional[str] = attr.ib(kw_only=True, default=None)
    api_key: Optional[str] = attr.ib(kw_only=True, default=None)
    token: Optional[str] = attr.ib(kw_only=True, default=None)


@attr.s(slots=True, kw_only=True)
class PlanConfig:
    """Test plan configuration."""

    # Current config version
    version = CONFIG_VERSION

    target_url: str = attr.ib(kw_only=True)
    spec_path: Optional[str] = attr.ib(kw_only=True, default=None)
    test_plan_path: Optional[str] = attr.ib(kw_only=True, default=None)
    # This is deprecated and should be removed in next version.
    auth: Optional[Tuple[str, str]] = attr.ib(kw_only=True, default=None)
    auth_type: Optional[str] = attr.ib(kw_only=True, default=None)
    report_to_saas: bool = attr.ib(kw_only=True, default=True)
    auth_config: Optional[AuthConfig] = attr.ib(kw_only=True, default=None)
    headers: Dict[str, str] = attr.ib(kw_only=True, factory=dict)
    env_file_path: Optional[str] = attr.ib(kw_only=True, default=None)
    ignore_ssl_verify: bool = attr.ib(kw_only=True, default=False)
    suite_execution_delay: int = attr.ib(kw_only=True, default=0)
    case_execution_delay: int = attr.ib(kw_only=True, default=0)
    request_timeout: int = attr.ib(kw_only=True, default=DEFAULT_REQUEST_TIMEOUT)

    # Run tests for these HTTP methods only
    http_methods: Optional[List[str]] = attr.ib(kw_only=True, default=None)

    # Module providers
    module_providers: Dict[Module, Provider] = attr.ib(kw_only=True, factory=dict)

    # Log handlers
    test_case_log_handler: Optional[Handler] = attr.ib(kw_only=True, default=None)
    runner_log_handler: Optional[Handler] = attr.ib(kw_only=True, default=None)

    # Shortcut to convert PlanConfig to dictionary
    asdict = attr.asdict



#
#  Copyright ©2024. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#
import base64
import json
from typing import Optional, Tuple, Union

import requests

from .models import Response

# These are the keys that we'll be looking for inside the body to get the error code.
ERROR_CODE_KEYS = [
    "code",
    "status",
    "statusCode",
    "errorCode",
    "error_code",
    "status_code",
    "error",
]

STRING_CODE_TO_STATUS_DICT = {
    "unknown": 400,
    "invalid_input": 400,
    "param_should_be_present": 400,
    "failed": 400,
    "no_exchange_enabled": 400,
    "unauthorized": 401,
    "user_does_not_exist": 404,
    "retry": 503,
    "wrong_params": 400,
    "token_auth_fail": 401,
    "invalid_token": 401,
    "forbidden": 403,
}


# pylint: disable=too-many-branches
def get_error_code_from_response_body(
    response: Union[str, bytes, Response, requests.Response]
) -> Optional[int]:
    """Check if the response body contains an error code and return it.

    Integer error codes take precedence over string error codes.
    """
    body_str = (
        response
        if isinstance(response, (str, bytes))
        else response.content
        if isinstance(response, requests.Response)
        else response.body
    ) or ""
    # TODO: Be consistent about base64 encoding of response body.
    try:
        body = json.loads(body_str)
    except Exception:
        try:
            decoded_body = base64.b64decode(body_str)
            encoding = "utf8"
            if not isinstance(response, (str, bytes)) and response.encoding:
                encoding = response.encoding
            response_body = decoded_body.decode(encoding, errors="replace")
            body = json.loads(response_body)
        except Exception:
            return None

    if not isinstance(body, dict):
        return None

    found_string_error_code = None

    for key in ERROR_CODE_KEYS:
        if value := body.get(key):
            if isinstance(value, int) or (isinstance(value, str) and value.isdigit()):
                status_code = int(value)
                if 100 <= status_code < 600:
                    return status_code
            elif (
                not found_string_error_code
                and isinstance(value, str)
                and value.casefold() in STRING_CODE_TO_STATUS_DICT
            ):
                found_string_error_code = STRING_CODE_TO_STATUS_DICT[value.casefold()]

    error_dict = {}
    # TODO: we should configure error dict's jsonpath per customer
    if (
        "error" in body
        and body["error"] is not None
        and isinstance(body["error"], dict)
    ):
        error_dict = body["error"]
    elif body.get("data") and isinstance(body["data"], dict):
        error_dict = body["data"]

    if error_dict:
        for key in ERROR_CODE_KEYS:
            if value := error_dict.get(key):
                if isinstance(value, int) or (
                    isinstance(value, str) and value.isdigit()
                ):
                    status_code = int(value)
                    if 100 <= status_code < 600:
                        return status_code
                elif (
                    not found_string_error_code
                    and isinstance(value, str)
                    and value.casefold() in STRING_CODE_TO_STATUS_DICT
                ):
                    found_string_error_code = STRING_CODE_TO_STATUS_DICT[
                        value.casefold()
                    ]
                    return found_string_error_code

    return found_string_error_code


def is_valid_2xx_response(
    response: Union[Response, requests.Response], return_code: bool = False
) -> Union[bool, Tuple[bool, int]]:
    """Check if the response is a valid 2xx response.

    In addition to checking the status code, we also check if the response body doesn't have any error codes because
    some APIs return 200 OK for error responses but the response body contains an error code.
    """
    if isinstance(response, requests.Response):
        response = Response.from_requests(response)

    # Check if the response is an error response.
    if response.status_code < 200 or response.status_code >= 300:
        return False if not return_code else (False, response.status_code)

    if response.body is not None:
        try:
            code = get_error_code_from_response_body(response)

            # If the code is present in the body, we have to return based on that.
            if code:
                return (
                    200 <= code < 300 if not return_code else (200 <= code < 300, code)
                )

        except Exception:
            pass
    return True if not return_code else (True, response.status_code)


def is_4xx_error_response(
    response: Union[Response, requests.Response], return_code: bool = False
) -> Union[bool, Tuple[bool, int]]:
    if isinstance(response, requests.Response):
        response = Response.from_requests(response)

    # Check if the response is an error response.
    if 400 <= response.status_code < 500:
        return True if not return_code else (True, response.status_code)

    if 200 <= response.status_code < 300 and response.body is not None:
        # Some APIs return 200 OK for error responses but the response body contains an error code.
        # So, we check if the response body contains an error code.
        try:
            code = get_error_code_from_response_body(response)

            # If the code is defined in the body, return based on that.
            if code:
                return (
                    400 <= code < 500 if not return_code else (400 <= code < 500, code)
                )
        except Exception:
            pass
    return False if not return_code else (False, response.status_code)



#
#  Copyright ©2022. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#

from levo_commons.config import AuthConfig, PlanConfig


def test_default_config():
    config = PlanConfig(
        target_url="http://localhost",
        headers={"myheader": "myvalue"},
        auth_config=AuthConfig(auth_type="basic", username="test", password="test"),
    )
    assert len(config.headers) == 1
    assert config.asdict() == {
        "target_url": "http://localhost",
        "spec_path": None,
        "test_plan_path": None,
        "auth": None,
        "auth_type": None,
        "report_to_saas": True,
        "env_file_path": None,
        "headers": {"myheader": "myvalue"},
        "auth_config": {
            "auth_type": "basic",
            "username": "test",
            "password": "test",
            "api_key": None,
            "token": None,
        },
        "module_providers": {},
        "runner_log_handler": None,
        "test_case_log_handler": None,
        "ignore_ssl_verify": False,
        "suite_execution_delay": 0,
        "case_execution_delay": 0,
        "request_timeout": 30,
        "http_methods": None,
    }


def test_config_with_overrides():
    config = PlanConfig(
        target_url="http://localhost:8080",
        spec_path="/my/local/path",
        headers={},
        auth_config=None,
        report_to_saas=False,
        ignore_ssl_verify=True,
        suite_execution_delay=10,
        case_execution_delay=10,
        request_timeout=20,
        http_methods=["GET"],
    )
    assert len(config.headers) == 0
    assert config.asdict() == {
        "target_url": "http://localhost:8080",
        "spec_path": "/my/local/path",
        "test_plan_path": None,
        "auth": None,
        "auth_type": None,
        "report_to_saas": False,
        "env_file_path": None,
        "headers": {},
        "auth_config": None,
        "module_providers": {},
        "runner_log_handler": None,
        "test_case_log_handler": None,
        "ignore_ssl_verify": True,
        "suite_execution_delay": 10,
        "case_execution_delay": 10,
        "request_timeout": 20,
        "http_methods": ["GET"],
    }



#
#  Copyright ©2022. Levo.ai Inc. All Rights Reserved.
#  You may not copy, reproduce, distribute, publish, display, perform, modify, create derivative works, transmit,
#  or in any way exploit any such software/code, nor may you distribute any part of this software/code over any network,
#  including a local area network, sell or offer it for commercial purposes.
#

from . import config, events, failures, models, providers, status, types, utils
